#!/usr/bin/python
# ConfigInfo.py -- Read the configuration file and record the configuration information

import os
import ConfigParser
from shutil import copyfile
from gaft.mpiutil import mpi
from GenQuery import updateVeriCmd
from ExcelOutput import initExcel


class ConfigObj(object):
    '''
    Class for reading the configuration information.
    '''

    def __init__(self, configfilename):
        # Read configuration file
        config = ConfigParser.ConfigParser()
        config.readfp(open(configfilename))

        # Initialize the configuration variables
        self.UppaalPath = config.get("MODEL", "UppaalPath")
        self.UppaalVeriCmd = config.get("MODEL", "VeriCmd")
        self.UppaalObjCmd = config.get("MODEL", "ObjCmd")
        self.UppaalMCCmd = config.get("MODEL", "MCCmd")
        self.ModelFileName = config.get("MODEL", "ModelFile")
        self.VeriFileName = config.get("MODEL", "VeriFile")
        self.PartNum = int(config.get("MODEL", "PartitionNumber"))
        self.UnitScale = float(config.get("MODEL", "UnitScale"))
        self.MinPrd = int(config.get("PARAM", "MinPeriod"))
        self.MaxPrd = int(config.get("PARAM", "MaxPeriod"))
        self.MaxMajorFrame = int(config.get("PARAM", "MaxMajorFrame"))
        self.TimeBound = int(config.get("SEARCH", "TimeBound"))
        self.PrUB = float(config.get("SEARCH", "PrUpperBound"))
        self.Step = int(config.get("SEARCH", "Step"))
        self.PrnCmd = config.get("SEARCH", "PruningCmd")
        self.IterNum = int(config.get("GA", "Iteration"))
        self.PoplSize = int(config.get("GA", "PopulationSize"))
        self.ElitistSize = int(config.get("GA", "ElitistSize"))
        self.FitFactor = int(config.get("GA", "FitFactor"))
        self.CrossPro = float(config.get("GA", "CrossPro"))
        self.ExchPro = float(config.get("GA", "ExchangePro"))
        self.LinearDelta = float(config.get("GA", "LinearDelta"))
        self.MutatePro = float(config.get("GA", "MutatePro"))
        self.MutateSigma = int(config.get("GA", "MutateSigma"))
        self.VeriGen = int(config.get("GA", "VeriGenerations"))
        self.MinStrategy1 = int(config.get("Strategy", "MinSigma1"))
        self.MaxStrategy1 = int(config.get("Strategy", "MaxSigma1"))
        self.MinStrategy2 = int(config.get("Strategy", "MinSigma2"))
        self.MaxStrategy2 = int(config.get("Strategy", "MaxSigma2"))
        self.Strategy1 = int(config.get("Strategy", "Sigma1"))
        self.Strategy2 = int(config.get("Strategy", "Sigma2"))
        self.SMLRate = float(config.get("Strategy", "MutateLearningRate"))
        self.SRLRate = float(config.get("Strategy", "CrossLearningRate"))
        self.SDiscountRate = float(config.get("Strategy", "DiscountRate"))
        self.SActionStrength = float(config.get("Strategy", "ActionStrength"))
        self.SProgressGen = int(config.get("Strategy", "ProgressGen"))
        self.SLGen = int(config.get("Strategy", "LearningGen"))
        self.PoplFile = config.get("GA", "PopulationFile")
        try:
            self.PoplFp = open(self.PoplFile, "r+")
        except IOError:
            self.PoplFp = None

        # Change the current path
        os.chdir(self.UppaalPath)

        # Open UPPAAL model file
        copyfn = '_' + str(mpi.rank) + '_' + self.ModelFileName
        copyfile(self.ModelFileName, copyfn)
        self.ModelFileName = copyfn
        self.Fp = open(self.ModelFileName, 'r+')
        self.Content = self.Fp.readlines()

        # Update the verification commands with the new model file name
        self.UppaalVeriCmd = updateVeriCmd(self.UppaalVeriCmd, self.ModelFileName)
        self.UppaalObjCmd = updateVeriCmd(self.UppaalObjCmd, self.ModelFileName)
        self.UppaalMCCmd = updateVeriCmd(self.UppaalMCCmd, self.ModelFileName)
        self.PrnCmd = updateVeriCmd(self.PrnCmd, self.ModelFileName)

        # Open output Excel file and initialize its format
        self.Workbook = initExcel(self.PartNum)


class StatusObj(object):
    '''
    Class for recording the optimization result information.
    '''

    def __init__(self, cfg):
        self.Prd = [cfg.MaxPrd for i in range(cfg.PartNum)]
        self.Budg = [cfg.Step for j in range(cfg.PartNum)]
        self.BestPrd = [cfg.MaxPrd for k in range(cfg.PartNum)]
        self.BestBudg = [cfg.Step for l in range(cfg.PartNum)]
        self.BestUtil = 0
        self.BestRes = 0
        self.DataCnt = 0
        self.DataRow = 0
