#!/usr/bin/python
# GenModel.py -- Generate UPPAAL model files

import os
import re


descstr1 = "const mf_t MajorFrame = "
descstr2 = "const int TotalWindowsNum = "
descstr3 = "const partition_window_t partition_windows[TotalWindowsNum] = {"
descstr4 = "const partition_t partition[Partitions] = {"


# getNextWindow -- To solve the start of task's next partition window
#  in the next partition period after task's initial offset
def getNextWindow(tskioffset, pno, pprd, plist, majorframe):
    n_m = tskioffset / majorframe
    r_m = tskioffset % majorframe
    n_p = r_m / pprd
    r_p = r_m % pprd
    if r_p == 0:
        o_m = r_m
    else:
        o_m = (n_p + 1) * pprd
        if o_m >= majorframe:
            n_m += 1
            o_m = 0

    for win in plist:
        if win[0] == pno and win[1] >= o_m:
            o_m = win[1]
            break

    return n_m * majorframe + o_m


# Type of message information
class MessageInfo(object):
    """Item type of a message"""
    def __init__(self, initoffset=0, period=0, offset=0, jitter=0):
        self.initoffset = initoffset
        self.period = period
        self.offset = offset
        self.jitter = jitter

    def init(self, msgno, content, pno, pprd, plist, majorframe):
        state = 0
        STATE_FIND_TASK_LINE = 1
        STATE_GET_TASK_LINE = 2
        STATE_GET_DATA_LINE = 3
        tskstr = ""
        offset = 0

        for s in content:
            # Enter different states
            if "const operation_t op[OperationTableLength] = {" in s:
                state = STATE_FIND_TASK_LINE
            elif "const task_t task[Tasks] = {" in s:
                state = STATE_GET_TASK_LINE

            # Execution based on the current state
            if state == STATE_FIND_TASK_LINE:
                matchobj = re.match(r'\s+// (Task\d+)', s)
                if matchobj:
                    tskstr = matchobj.group(1)
                    offset = 0
                else:
                    matchobj = re.match(r'\s+\{COMPUTE,\s*CPU,\s*(\d+),', s)
                    if matchobj:
                        offset += int(matchobj.group(1))
                    else:
                        matchobj = re.match(r'\s+\{SEND,\s*MSG' + str(msgno + 1), s)
                        if matchobj:
                            state = STATE_GET_TASK_LINE
            elif state == STATE_GET_TASK_LINE:
                if re.match(r'\s+// ' + tskstr, s):
                    state = STATE_GET_DATA_LINE
            elif state == STATE_GET_DATA_LINE:
                matchobj = re.match(r'\s+\{Periodic,\s*(\d+),\s*(\d+),\s*(\d+),\s*(\d+),\s*(\d+),\s*(\d+),', s)
                self.initoffset = getNextWindow(int(matchobj.group(1)), pno, pprd, plist, majorframe)
                self.period = int(matchobj.group(2))
                self.offset = int(matchobj.group(4)) + offset - 1
                self.jitter = int(matchobj.group(6)) - self.offset + 1
                state = 0
                break

        return self


# genMFStr -- Generate the string of major frame's declaration in the UPPAAL model
def genMFStr(mf):
    return descstr1 + str(mf) + ";\n"


# genWNumStr -- Generate the string of number of partition windows' declaration in the UPPAAL model
def genWNumStr(n):
    return descstr2 + str(n) + ";\n"


# writeScheduleStr -- Generate the string of schedule's declaration in the UPPAAL model
def writeScheduleStr(fp, plist, pindex):
    fp.writelines(descstr3 + "\n")

    i = j = 0
    for (p, o, b) in plist:
        if i == pindex[j]:
            fp.writelines("\t// Partition " + str(j + 1) + "\n")
            if j < len(pindex) - 1:
                j += 1
        if i < len(plist) - 1:
            s = "\t{" + str(o) + ", " + str(b) + "},\n"
        else:
            s = "\t{" + str(o) + ", " + str(b) + "}\n"
        fp.writelines(s)
        i += 1

    fp.writelines("};\n")

    return


# genPartStr -- Generate the string of partitions' declaration in the UPPAAL model
def genPartStr(line, period, wcnt, index):
    matchobj = re.match(r'(\s*\{(\w+,\s*){2})', line)
    if matchobj:
        s = matchobj.group(1)
        s = s + str(period) + ", " + str(wcnt) + ", " + str(index) + "}"
    else:
        s = "};"

    return s


# buildModelFile -- Build the complete UPPAAL model file
def buildModelFile(fp, content, mf, plist, pindex, pprd):

    fp.seek(0)
    fp.truncate(0)

    pnum = len(pindex)
    wcnts = []
    for i in range(pnum):
        if i < pnum - 1:
            wcnts.append(pindex[i + 1] - pindex[i])
        else:
            wcnts.append(len(plist) - pindex[i])

    state = 0
    for i in range(len(content)):
        if re.match(descstr1, content[i]):
            state = 1
        if re.match(descstr2, content[i]):
            state = 2
        if content[i].startswith(descstr3, 0, len(descstr3)):
            state = 3
        if content[i].startswith(descstr4, 0, len(descstr4)):
            state = 4

        if state == 0:
            fp.writelines(content[i])
        if state == 1:
            fp.writelines(genMFStr(mf))
            state = 0
        if state == 2:
            fp.writelines(genWNumStr(len(plist)))
            state = 0
        if state == 3:
            if re.match(r'\};', content[i]):
                writeScheduleStr(fp, plist, pindex)
                state = 0
        if state == 4:
            fp.writelines(content[i])
            state = 5
        elif state >= 5:
            fp.writelines(genPartStr(content[i], pprd[state - 5], wcnts[state - 5], pindex[state - 5]))
            if state - 5 < pnum - 1:
                fp.writelines(",")
                state += 1
            fp.writelines("\n")
            if re.match(r'\};', content[i]):
                state = 0

    fp.flush()
    os.fsync(fp)

    return


# buildMIFile -- Build the UPPAAL model file for the construction of message interfaces
def buildMIFile(fp, content, msg, initoffset, period, offset, jitter):

    fp.seek(0)
    fp.truncate(0)

    state = 0
    STATE_GLOBAL_ANALYSIS_LINE = 1
    STATE_COMPOS_ANALYSIS_LINE = 2
    STATE_COMPOS_PORT_LINE = 3
    STATE_MSG_LINE = 4

    new = []
    template_lines = []
    cur_pno = 0
    pno = 0
    enable = False

    # Uncomment template declaration
    for i in range(len(content)):
        # Enter different states
        if "// Global analysis" in content[i]:
            state = STATE_GLOBAL_ANALYSIS_LINE
        elif "// Compositional analysis" in content[i]:
            state = STATE_COMPOS_ANALYSIS_LINE

        # Execution based on the current state
        if state == STATE_GLOBAL_ANALYSIS_LINE:
            if content[i].startswith("//"):
                new.append(content[i])
            else:
                new.append("//" + content[i])
        elif state == STATE_COMPOS_ANALYSIS_LINE:
            matchobj = re.match(r'// Partition (\d+)', content[i])
            if matchobj:
                cur_pno = int(matchobj.group(1))
                for s in template_lines:
                    if enable:
                        new.append(s.replace("//", ""))
                    else:
                        new.append("//" + s.replace("//", ""))

                enable = False
                template_lines = []
                new.append(content[i])
            elif re.match(r'\s+', content[i]):
                for s in template_lines:
                    if enable:
                        new.append(s.replace("//", ""))
                    else:
                        new.append("//" + s.replace("//", ""))

                enable = False
                template_lines = []
                new.append(content[i])
                state = 0
            else:
                if re.search(r'msg' + str(msg + 1) + '_rcv', content[i]):
                    pno = cur_pno
                    enable = True
                template_lines.append(content[i])
        else:
            new.append(content[i])

    enable = False
    moff = 0

    # Uncomment global declaration
    for s in new:
        # Enter different states
        if "// Ports disabled in global analysis" in s:
            state = STATE_COMPOS_PORT_LINE
        elif "const msg_t msg[MSGs] = {" in s:
            state = STATE_MSG_LINE
        elif re.match(r'^[\s]$', s):
            state = 0

        # Execution based on the current state
        if state == STATE_COMPOS_PORT_LINE:
            if enable:
                fp.writelines(s.replace("//", ""))
                enable = False
            else:
                fp.writelines("//" + s.replace("//", ""))
            if "// Ports disabled in compositional analysis for partition " + str(pno) in s:
                enable = True
        elif state == STATE_MSG_LINE:
            if moff == msg + 1:
                matchobj = re.match(r'(\s*\{MSG\d+,\s)(\d+,\s)+(.*)', s, re.S)
                if matchobj:
                    fp.writelines(matchobj.group(1) + str(initoffset) + ", " + str(period) + ", "
                                  + str(period) + ", " + str(offset) + ", " + str(jitter) + ", " + matchobj.group(3))
                else:
                    print "MSG format error in UPPAAL template file!"
                    fp.writelines(s)
            else:
                fp.writelines(s)
            moff += 1
        else:
            fp.writelines(s)

    fp.flush()
    os.fsync(fp)

    return


# getMsgNum -- Get the number of message number from the UPPAAL model file
def getMsgNum(content):
    for s in content:
        matchobj = re.match(r'const int MSGs = (\d+);', s)
        if matchobj:
            return int(matchobj.group(1))


# getSrcPartition -- Get the source partition of message from the UPPAAL model file
def getSrcPartition(msgno, content):
    state = 0
    STATE_COMPOS_ANALYSIS_LINE = 1
    pno = cur_pno = -1

    for s in content:
        if "// Compositional analysis" in s:
            state = STATE_COMPOS_ANALYSIS_LINE

        if state == STATE_COMPOS_ANALYSIS_LINE:
            matchobj = re.match(r'// Partition (\d+)', s)
            if matchobj:
                cur_pno = int(matchobj.group(1)) - 1
            else:
                if re.search(r'msg' + str(msgno + 1) + '_rcv', s):
                    pno = cur_pno

    if pno < 0:
        print "Cannot obtain source partition from message type!"

    return pno


# buildCompVeriModelFile -- Build the UPPAAL model file for the compositional verification of partition pno
def buildCompVeriModelFile(fp, content, pno):

    fp.seek(0)
    fp.truncate(0)

    state = 0
    STATE_COMPOS_PORT_LINE = 1
    STATE_GLOBAL_ANALYSIS_LINE = 2
    STATE_COMPOS_ANALYSIS_LINE = 3
    enable = False
    cur_pno = -1
    template_lines = []

    for s in content:
        # Enter different states
        if "// Ports disabled in global analysis" in s:
            state = STATE_COMPOS_PORT_LINE
        elif "// Global analysis" in s:
            state = STATE_GLOBAL_ANALYSIS_LINE
        elif "// Compositional analysis" in s:
            state = STATE_COMPOS_ANALYSIS_LINE

        # Execution based on the current state
        if state == STATE_COMPOS_PORT_LINE:
            if re.match(r'^[\s]$', s):
                enable = True
                state = 0

            if enable:
                fp.writelines(s.replace("//", ""))
                enable = False
            else:
                fp.writelines("//" + s.replace("//", ""))

            if "// Ports disabled in compositional analysis for partition " + str(pno + 1) in s:
                enable = True
        elif state == STATE_GLOBAL_ANALYSIS_LINE:
            if s.startswith("//"):
                fp.writelines(s)
            else:
                fp.writelines("//" + s)
        elif state == STATE_COMPOS_ANALYSIS_LINE:
            matchobj = re.match(r'// Partition (\d+)', s)
            if matchobj:
                for ss in template_lines:
                    if enable:
                        fp.writelines(ss.replace("//", ""))
                    else:
                        fp.writelines("//" + ss.replace("//", ""))

                cur_pno = int(matchobj.group(1)) - 1
                if cur_pno == pno:
                    enable = True
                else:
                    enable = False

                template_lines = []
                fp.writelines(s)
            elif re.match(r'^[\s]$', s):
                for ss in template_lines:
                    if enable:
                        fp.writelines(ss.replace("//", ""))
                    else:
                        fp.writelines("//" + ss.replace("//", ""))

                enable = False
                template_lines = []
                fp.writelines(s)
                state = 0
            else:
                template_lines.append(s)
        else:
            fp.writelines(s)

    fp.flush()
    os.fsync(fp)

    return


# outputMI -- Output message interfaces to the UPPAAL model text
def outputMI(content, msglist):
    for i in range(len(content)):
        matchobj = re.match(r'(\s*\{MSG(\d+),\s)(\d+,\s)+(.*)', content[i], re.S)
        if matchobj:
            n = int(matchobj.group(2)) - 1
            content[i] = matchobj.group(1) + str(msglist[n].initoffset) + ", " + str(msglist[n].period) + ", "\
                         + str(msglist[n].period) + ", " + str(msglist[n].offset) + ", " + str(msglist[n].jitter)\
                         + ", " + matchobj.group(4)


# getMajorFrame -- Get major frame from model file text
def getMajorFrame(content):
    for s in content:
        matchobj = re.match(r'' + descstr1 +'(\d+);', s)
        if matchobj:
            return int(matchobj.group(1))
