#!/usr/bin/python
# Optim_es.py -- Optimization by exhaustive search

# Import system functions
import os
import re
from math import ceil

# Import user-defined functions
from GenModel import genParamStr
from GenModel import buildModelFile
from GenQuery import genVeriStr
from GenQuery import genVeriFile
from GenQuery import genPruningStr
from GenQuery import pruningCmd
from GenQuery import genPruningFile
from ExcelOutput import outputExcel


# runVeri -- Perform schedulability verification under current parameters
def runVeri(cmd):
    resstr = os.popen(cmd)
    info = resstr.readlines()
    normal = False
    for line in info:
        line = line.strip('\r\n')
        if re.search(r'-- Formula is ', line):
            normal = True
            if re.search(r'NOT', line):
                result = False
            else:
                result = True   # Satisfaction(bool type)
        if re.match(r'\([\d]+ runs\) Pr', line):
            result = float(re.search(r',([0]*\.*\d*)', line).group(1))  # Error probability(float type)
        if re.match(r'\([\d]+ runs\) E', line):
            result = int(re.search(r'[-]?\d+$', line).group())  # Utilization(int type)

    if normal:
        return result
    else:
        print "UPPAAL command is invalid!"
        exit(1)


# promptParam -- Prompt users to monitor the parameters
def promptParam(cfg, stat):
    print "Case No." + str(stat.DataCnt + 1)
    print "Test parameter:",
    for i in range(cfg.PartNum):
        if i == 0:
            print "Partition_%d period = %d budget = %d" % (i + 1, stat.Prd[i], stat.Budg[i]),
        else:
            print "\n\t\t\t\tPartition_%d period = %d budget = %d" % (i + 1, stat.Prd[i], stat.Budg[i]),
    print ''


# updateResult -- Update the result and output the information
def updateResult(issched, cfg, stat):
    stat.DataCnt += 1
    # Case 1: Schedulable
    if issched:
        # Run UPPAAL to calculate the utilization
        util = runVeri(cfg.UppaalObjCmd)
        outputExcel(cfg.Workbook, cfg.PartNum, stat.DataCnt, stat.Prd, stat.Budg, 'Y', util / 100.0, cfg.UnitScale)
        stat.DataRow += 1
        if util > stat.BestUtil:
            stat.BestPrd = list(stat.Prd)
            stat.BestBudg = list(stat.Budg)
            stat.BestUtil = util
            stat.BestRes = stat.DataRow
        print "Result: Schedulable\n\t\tUtilization = %.2f" % (util / 100.0)
        ret = 1
    # Case 2: Non-schedulable
    else:
        #outputExcel(wb, partnum, DataCnt, Prd, Budg, 'N', 0.0, UnitScale)
        print "Result: Non-schedulable\n\t\tUtilization = NONE"
        ret = 0

    # Output the best result up to now
    if stat.BestUtil != 0:
        print "Current result:",
        for i in range(cfg.PartNum):
            if i == 0:
                print "P%d period = %d budget = %d" % (i + 1, stat.BestPrd[i], stat.BestBudg[i]),
            else:
                print "\n\t\t\t\tP%d period = %d budget = %d" % (i + 1, stat.BestPrd[i], stat.BestBudg[i]),
        print "\t--->\tUtilization = %.2f" % (stat.BestUtil / 100.0)
    else:
        print "Current result: NONE"
    print "------------------------------------------------"

    return ret


# printPruningInf -- Show the pruned branch
def printPruningInf(prd, budg, layer):

    for i in range(layer + 1):
        if i == 0:
            print "Prune parameter:Partition_%d period = %d budget = %d" \
                  % (i + 1, prd[i], budg[i]),
        else:
            print "\n\t\t\t\tPartition_%d period = %d budget = %d" % (i + 1, prd[i], budg[i]),
    print "\n------------------------------------------------"


#bRange -- Get the range of budget
def bRange(budgid, prdi, step, stat, scale):
    rate = scale
    for i in range(budgid):
        rate -= stat.Budg[i] * scale / stat.Prd[i]
    if rate < 0 or rate >= scale:
        return range(step, prdi, step)
    else:
        return range(step, int(ceil(float(prdi * rate) / scale)), step)


# paramSweep_ES -- Perform recursive parameter sweep by exhaustive search
def _paramSweep(start, end, cfg, stat):

    if start == end:
        # Prompt users
        promptParam(cfg, stat)
        # Modify UPPAAL model file
        buildModelFile(cfg.Fp, cfg.Content, cfg.Paramlno, genParamStr(stat.Prd, stat.Budg, cfg.PartNum))
        # Run UPPAAL to verify the schedulability
        res = runVeri(cfg.UppaalVeriCmd)
        # Update the result and output the information
        return updateResult(res, cfg, stat)

    ret = 0
    for prdi in range(cfg.MaxPrd, cfg.MinPrd - 1, 0 - cfg.Step):
        for budgi in bRange(start, prdi, cfg.Step, stat, 100):
            stat.Prd[start] = prdi
            stat.Budg[start] = budgi

            # Pre-order Branch pruning
            if start < end - 1:
                # Modify UPPAAL model file
                buildModelFile(cfg.Fp, cfg.Content, cfg.Paramlno, genParamStr(stat.Prd, stat.Budg, cfg.PartNum))
                # Run UPPAAL to verify the schedulability of partition "start"
                res = runVeri(pruningCmd(cfg.PrnCmd, start))
                # Prune the current branch and output the information
                if not res:
                    printPruningInf(stat.Prd, stat.Budg, start)
                    continue

            # Recursive invocation
            hassched = _paramSweep(start + 1, end, cfg, stat)

            # Post-order Branch pruning
            if hassched:
                ret = hassched  # Renew the return value
                break
            if not hassched and start < end - 1:
                break

    return ret


# paramSweep_ES -- Parameter sweep by exhaustive search
def paramSweep_ES(cfg, stat):
    # Generate the pruning query file for each partition
    genPruningFile(cfg.PrnCmd, genPruningStr(cfg.Content, cfg.PartNum, cfg.TimeBound, cfg.PrUB), cfg.PartNum)
    # Generate the verification query file
    genVeriFile(cfg.VeriFileName, genVeriStr(cfg.TimeBound, cfg.PrUB))

    # Run parameter sweep
    _paramSweep(0, cfg.PartNum, cfg, stat)
