#!/usr/bin/python
# Optim_ga.py -- Optimization by genetic algorithm

# Import system functions
import re
from gaft.mpiutil import mpi
import logging
from sys import stdout
from random import random
from bisect import bisect_right

# Import gaft functions
from gaft import GAEngine
from gaft.components import GAIndividual
from gaft.components import GAPopulation
from gaft.operators import RouletteWheelSelection
from gaft.operators import TruncationSelection
from gaft.operators import TournamentSelection
from gaft.operators import LinearRankingSelection
from gaft.operators import ExponentialRankingSelection
from gaft.operators import LogarithmicRankingSelection
from gaft.operators import UniformCrossover
from gaft.operators import GroupUniformCrossover
from gaft.operators import IntermediateRecombination
from gaft.operators import LineRecombination
from gaft.operators import LocalLineRecombination
from gaft.operators import RadianLineRecombination
from gaft.operators import DeltaMutation
from gaft.operators import RotatingDeltaMutation
from gaft.operators import RotatingGaussMutation
from gaft.operators import FlipBitMutation
from gaft.analysis.fitness_store import FitnessStoreAnalysis
from gaft.plugin_interfaces.analysis import OnTheFlyAnalysis

# Import user-defined functions
from GenModel import buildModelFile
from GenQuery import genGaVeriStr
from GenQuery import genGaVeriFile
from GenQuery import gaVeriCmd
from ExcelOutput import outputExcel
from GenSchedule import leastCommonMultiple
from GenSchedule import genSchedule
from GenSchedule import genUtil
from Verification import runVeri
from Verification import composVeri


# promptParam -- Prompt users to monitor the parameters
def promptStep(g, population, engine):
    best_indv = population.best_indv(engine.fitness)
    msg = 'Generation: {}, best fitness: {:.3f}'.format(g, engine.fitness(best_indv))
    for indv in population.individuals:
        msg += '\n'
        for i in range(len(indv.variants)):
            msg += str(indv.variants[i]) + ' '
        msg += str(indv.lasting) + '('
        for i in range(len(indv.strategy)):
            msg += str(int(indv.strategy[i])) + ' '
        msg += ')\t\t' + str(engine.fitness(indv))
        pass
    msg += '\n-----------------------------------------------------------------------'
    return msg


# updateResult -- Update the result and output the information to an excel file
def updateResult(cfg, stat, g, population, engine):
    for indv in population.individuals:
        stat.DataCnt += 1
        fit = engine.fitness(indv)
        stat.Prd = [indv.variants[2 * i] for i in range(cfg.PartNum)]
        stat.Budg = [indv.variants[2 * i + 1] for i in range(cfg.PartNum)]
        # Schedulable case
        if fit >= (cfg.PartNum + 1) * cfg.FitFactor:
            stat.DataRow += 1
            util = (fit - (cfg.PartNum + 1) * cfg.FitFactor) * 10000.0 / cfg.FitFactor
            outputExcel(cfg.Workbook, cfg.PartNum, g, stat.Prd, stat.Budg, 'Y', util / 100.0, cfg.UnitScale)
            if util > stat.BestUtil:
                stat.BestPrd = list(stat.Prd)
                stat.BestBudg = list(stat.Budg)
                stat.BestUtil = util
                stat.BestRes = stat.DataRow
        # Nonschedulable case
        else:
            stat.DataRow += 1
            outputExcel(cfg.Workbook, cfg.PartNum, g, stat.Prd, stat.Budg, 'N', 0.0, cfg.UnitScale)


class StrategySelection(object):
    '''
    Selection operator for strategy parameters
    '''
    def __init__(self, lrate=0.7, drate=0.5, arate=0.1, lg=1, maxlg=100, sranges=[], sinit=(), factor=100):
        # Learning rate of the strategy parameters (default value 0.7)
        self.lrate = lrate
        # Discount rate of the strategy parameters (default value 0.5)
        self.drate = drate
        # Action strength rate of the strategy parameters (default value 0.1)
        self.arate = arate
        # Generation gap in reinforcement learning
        self.lg = lg
        # Maximum generation in reinforcement learning
        self.maxlg = maxlg
        # Ranges of strategy parameters[(min1, max1),...]
        self.sranges = sranges
        # Init vaule of strategy parameters
        self.sinit = sinit
        # Factor of fitness value
        self.factor = factor
        # Reinitialize strategy parameters
        self.reinit = False

    def select(self, g, population, fitness):
        '''
        Select best strategy parameters using truncation selection.
        '''
        sortedindvs = sorted(population.individuals, key=fitness, reverse=True)
        bstf = population.max(fitness)

        selectedub = len(sortedindvs) / 2
        sortedindvs = sortedindvs[:selectedub]

        strtgy = [0 for i in range(len(sortedindvs[0].strategy))]
        for indv in sortedindvs:
            for i, ins in enumerate(indv.strategy):
                strtgy[i] += ins * (1 - self.lrate) ** indv.lasting

        strtgy = [si / selectedub for si in strtgy]

        # Check the range of strategy parameters
        if len(self.sranges) != 0:
            for i in range(len(strtgy)):
                if strtgy[i] < self.sranges[i][0]:
                    strtgy[i] = self.sranges[i][0]
                if strtgy[i] > self.sranges[i][1]:
                    strtgy[i] = self.sranges[i][1]

        self.keepInitStrategy(bstf, strtgy)
        self.reinitStrategy(strtgy)

        # Store the mean value only for output.
        population.meanstrtgy = [int(s) for s in strtgy]
        print 'The mean of strategy parameters: ', strtgy

        return strtgy

    def updateState(self, population, fitness):
        '''
        Update state in reinforcement learning
        '''
        # Find best fitness value in the current population
        bstf = population.max(fitness)

        # Update the best fitness value of the current population
        if bstf > population.bestfit:
            population.notinprogress = 0
        elif bstf <= population.bestfit:
            population.notinprogress += 1
        population.bestfit = bstf

        # Update the state of the current population
        if population.notinprogress == 0:
            population.state = 0
        elif 0 < population.notinprogress < self.lg:
            population.state = 1
        elif population.notinprogress >= self.lg:
            population.state = 2

    def keepInitStrategy(self, bstf, strtgy):
		'''
        Keep the initial strategy parameters to avoid the evolution of invalid information.
        '''
        #Find the lower bound
        keepbound = int(bstf - 1) / int(self.factor)

        for i in range(len(strtgy)):
            if i / 2 >= keepbound:
                strtgy[i] = self.sinit[i % 2]

    def reinitStrategy(self, strtgy):
		'''
        Keep a minimum mutation strength after the final convergence.
        '''
        if not self.reinit:
            reset = True
            if len(self.sranges) != 0:
                for i in range(len(strtgy)):
                    if i % 2 == 1:
                        if strtgy[i - 1] >= self.sranges[i - 1][0] + 5 or strtgy[i] >= self.sranges[i][0] + 1:
                            reset = False
            else:
                reset = False

            if reset:
                self.reinit = True
                for i in range(len(strtgy)):
                    if i % 2 == 0:
                        strtgy[i] = self.sranges[i + 1][0]
                    else:
                        strtgy[i] = self.sranges[i - 1][0]
                        self.sranges[i - 1][0] = self.sranges[i][0]

						
# initIndvsFromFile -- Initialize population from a file
def initIndvsFromFile(cfg, indv_template):
    if cfg.PoplFp is None:
        if mpi.is_master: print "Initialize Population Randomly..."
        return None

    if mpi.is_master: print "Initialize Population from File..."
    individuals = []
    fp = cfg.PoplFp
    info = fp.readlines()
    for line in info:
        line = line.strip('\n')
        sub1 = re.match(r'(\d+\s)+', line).group()
        sub2 = re.search(r'\(.+\)', line).group()
        sub3 = re.search(r'\)\s(\d+)', line)

        varstrs = re.findall(r'\d+\s', sub1)
        stystrs = re.findall(r'([\d\.]+)\s', sub2)
        variant = []
        strategy = []
        for var in varstrs:
            variant.append(int(var))
        for sty in stystrs:
            strategy.append(float(sty))

        indv = indv_template.clone()
        indv.init(variants=variant)
        indv.setStrategy(strategy)
        if sub3 is not None:
            indv.lasting = int(sub3.group(1))
        individuals.append(indv)

    fp.close()
    return individuals


# savePoplToFile -- Save population to a file on disks
def savePoplToFile(cfg, population):
    fp = open(cfg.PoplFile, 'w')
    for indv in population.individuals:
        for var in indv.variants:
            fp.write(str(var) + " ")
        fp.write("(")
        for sty in indv.strategy:
            fp.write(str(sty) + " ")
        fp.write(") " + str(indv.lasting) + "\n")

    fp.close()


# paramSweep_GA -- Perform recursive parameter sweep by genetic algorithm
def paramSweep_GA(cfg, stat):
    eps = cfg.Step
    minprd = cfg.MinPrd
    maxprd = cfg.MaxPrd

    # Generate the query file for each partition
    genGaVeriFile(cfg.VeriFileName, genGaVeriStr(cfg.TimeBound, cfg.PrUB, cfg.PartNum), cfg.PartNum)

    # Define population.
    ranges = []     # The ranges of object variables
    sranges = []    # The ranges of strategy parameters
    for i in range(cfg.PartNum):
        ranges.extend([(minprd, maxprd), (1, maxprd)])
        sranges.extend([[cfg.MinStrategy1, cfg.MaxStrategy1], [cfg.MinStrategy2, cfg.MaxStrategy2]])
        pass
    indv_template = GAIndividual(ranges=ranges, encoding='binary', eps=eps,
                                 strategy=[cfg.Strategy1, cfg.Strategy2] * cfg.PartNum)
    indvs = initIndvsFromFile(cfg, indv_template)
    population = GAPopulation(indv_template=indv_template, size=cfg.PoplSize, elitist_size=cfg.ElitistSize).init(indvs)

    # Create genetic operators.
    selection = ExponentialRankingSelection(base=cfg.CrossPro)
    crossover = LocalLineRecombination(d=cfg.LinearDelta)
    mutation = RotatingGaussMutation(pm=cfg.MutatePro, lrate=cfg.SMLRate, sranges=sranges)
    strategy_selection = StrategySelection(lrate=cfg.SRLRate, drate=cfg.SDiscountRate, arate=cfg.SActionStrength,
                                           lg=cfg.SProgressGen, maxlg=cfg.SLGen,
                                           sranges=sranges, sinit=(cfg.Strategy1, cfg.Strategy2),
                                           factor=cfg.FitFactor)

    # Create genetic algorithm engine.
    engine = GAEngine(cfg=cfg, population=population, selection=selection,
                      crossover=crossover, mutation=mutation,
                      analysis=[FitnessStoreAnalysis])

    # Define fitness function.
    def fitness(indv):
        factor = cfg.FitFactor
        partnum = cfg.PartNum
        verigen = cfg.VeriGen
        isverified = False

        exceed = 0.0
        rate = 0.0
        prds = [indv.variants[2 * i] for i in range(partnum)]
        budgs = [indv.variants[2 * i + 1] for i in range(partnum)]
        for i in range(partnum):
            prdi = prds[i]
            budgi = budgs[i]
            if budgi > prdi:
                exceed += (budgi - prdi)
            rate += (float(budgi) / prdi)
            pass
        if exceed > 0:
            totalbudg = sum(budgs)
            fit = -factor - factor * exceed / totalbudg
            return fit
        elif rate > 1.0:
            fit = -(rate - 1.0) / rate * factor
            return fit

        # Solve major frame of ARINC 653 schedule
        majorframe = leastCommonMultiple(prds)
        if majorframe > cfg.MaxMajorFrame:
            fit = 0.0
            return fit

        # Retrieve existing fitness value from dictionary
        if tuple(indv.variants) in engine.FitDict:
            fit = engine.FitDict[tuple(indv.variants)][0]
            indv.veri = isverified = engine.FitDict[tuple(indv.variants)][1]
            #print "Fetch in process", str(mpi.rank), indv.variants, fit, isverified
        # Calculate fitness value
        else:
            # Index of the first partition window for each partition
            pindex = []
            # List of the partition windows of partition schedule
            plist = []
            # Generate ARINC 653 schedule from periodic partition parameter vector
            schpcnt = genSchedule(majorframe, prds, budgs, pindex, plist)
            if schpcnt < partnum:
                fit = schpcnt * factor / partnum
            else:
                # Modify UPPAAL model file
                buildModelFile(cfg.Fp, cfg.Content, majorframe, plist, pindex, prds)
                # Run UPPAAL to verify the number of schedulable partitions
                engine.logger.info('Process ' + str(mpi.rank) + ' is verifying ' + str(indv.variants))
                res = [0 for k in range(partnum)]
                for i in range(partnum):
                    resi = runVeri(gaVeriCmd(cfg.UppaalVeriCmd, cfg.VeriFileName, i))
                    res[i] = resi[0]
                    if res[i] == 0:
                        break

                # Calculate fitness value based on utilization
                fit = sum(res) * factor + factor - genUtil(sum(res), majorframe, plist) * factor

        # Confirm schedulability by symbolic model checking
        if engine.gi >= verigen and partnum * factor < fit < (partnum + 1) * factor and not isverified:
            print "MC checking in process", str(mpi.rank), indv.variants
            #res = runVeri(cfg.UppaalMCCmd)
            pindex = []
            plist = []
            # Generate ARINC 653 schedule again
            genSchedule(majorframe, prds, budgs, pindex, plist)
            # Modify UPPAAL model file
            buildModelFile(cfg.Fp, cfg.Content, majorframe, plist, pindex, prds)
            cfg.Fp.seek(0)
            content_comp = cfg.Fp.readlines()
            # Run UPPAAL to perform compositional verification
            [res, schno] = composVeri(cfg.UppaalMCCmd, cfg.Fp, content_comp, partnum, prds, plist)
            indv.veri = isverified = True
            print "MC finishes checking in process", str(mpi.rank), indv.variants, res
            if res:
                fit = (partnum + 2) * factor - genUtil(partnum, majorframe, plist) * factor
            else:
                mfactor = factor / (partnum * 2)
                fit = partnum * factor + (schno + 1 - genUtil(partnum, majorframe, plist)) * mfactor

        #print "Result in Process ", str(mpi.rank), indv.variants, fit, indv.veri, isverified
        engine.FitDict[tuple(indv.variants)] = [fit, isverified]

        return fit

    # Register fitness function to the GA engine.
    engine.fitness_register(fitness)

    # Define on-the-fly analysis class.
    class ConsoleOutputAnalysis(OnTheFlyAnalysis):
        interval = 1
        master_only = True
        config = cfg
        status = stat

        def setup(self, ng, engine):
            generation_info = 'Generation number: {}'.format(ng)
            population_info = 'Population number: {}'.format(engine.population.size)
            self.logger.setLevel(logging.DEBUG)
            self.logger.addHandler(logging.StreamHandler(stdout))
            self.logger.info('Start GA ParamSweep...\n{} {}'.format(generation_info, population_info))

        def register_step(self, g, population, engine):
            self.logger.info(promptStep(g, population, engine))
            updateResult(self.config, self.status, g, population, engine)

        def finalize(self, population, engine):
            self.logger.info(promptStep(self.config.IterNum, population, engine))
            best_indv = population.best_indv(engine.fitness)
            x = best_indv.variants
            y = engine.fitness(best_indv)
            msg = 'Optimal solution: ({}, {})'.format(x, y)
            self.logger.info(msg)
            updateResult(self.config, self.status, self.config.IterNum, population, engine)
            savePoplToFile(self.config, population)

    # Register on-the-fly analysis to the GA engine.
    engine.analysis_register(ConsoleOutputAnalysis)

    # Register strategy selection function to the GA engine.
    engine.strategy_selection_register(strategy_selection)

    # Run the GA engine.
    engine.run(ng=cfg.IterNum)
