#!/usr/bin/python
# Verification.py -- Support global or compositional verification with UPPAAL

import os
import re
from gaft.mpiutil import mpi

from GenQuery import genVeriCmd
from GenModel import buildMIFile
from GenModel import getMsgNum
from GenModel import MessageInfo
from GenModel import getSrcPartition
from GenModel import buildCompVeriModelFile
from GenModel import outputMI
from GenModel import getMajorFrame

# runVeri -- Perform schedulability verification under current parameters
def runVeri(cmd):
    resstr = os.popen(cmd)
    info = resstr.readlines()
    normal = False
    satisfied = False
    result = []
    for line in info:
        line = line.strip('\n')
        if re.search(r'-- Formula ', line):
            normal = True
        if re.search(r'-- Formula is satisfied', line):
            satisfied = True
        if re.search(r'H0: Pr\(', line):
            result.append(1)  # The number of satisfaction(int type)
        if re.search(r'H1: Pr\(', line):
            result.append(0)  # The number of dissatisfaction(int type)
        if re.match(r'\([\d]+ runs\) Pr', line):
            result.append(float(re.search(r',([0]*\.*\d*)', line).group(1)))  # Error probability(float type)
        if re.match(r'\([\d]+ runs\) E', line):
            result.append(int(re.search(r'=[\s]+(([-]?\d+)(\.\d+)?)$', line).group(2)))  # Utilization(int type)

    if normal:
        if len(result):
            return result
        else:
            return satisfied
    else:
        print "UPPAAL command is invalid!"
        print cmd
        for line in info:
            print line
        exit(1)


# composVeri -- Perform compositional schedulability verification
def composVeri(cmd, fp, content_output, partnum, partprd, plist):
    # Get the number of message types in model file
    msgnum = getMsgNum(content_output)
    msglist = []
    res = False
    schno = 0

    # Check the satisfaction of deadlines of each partition
    for i in range(partnum):
        buildCompVeriModelFile(fp, content_output, i)
        print "\tProcess " + str(mpi.rank) + ": MC verifies the deadline of partition " + str(i + 1) + "..."
        res = runVeri(genVeriCmd(cmd, '', "query_ddl.txt"))
        if res:
            schno += 1
            print "\tProcess " + str(mpi.rank) + ": The deadline of partition " + str(i + 1) + " is satisfied!"
        else:
            print "\tProcess " + str(mpi.rank) + ": The deadline of partition " + str(i + 1) + " is missed!"
            break

    # Build the message interfaces for each partition
    if res:
        mf = getMajorFrame(content_output)
        for msgno in range(msgnum):
            cpno = getSrcPartition(msgno, content_output)
            msginfo = MessageInfo().init(msgno, content_output, cpno, partprd[cpno], plist, mf)
            print "\tProcess " + str(mpi.rank) + ": The initial message interface of MSG" + str(msgno + 1) + " is:\n" +\
                  "initialoffset: " + str(msginfo.initoffset) + " period: " + str(msginfo.period) + \
                  " offset: " + str(msginfo.offset) + " jitter: " + str(msginfo.jitter)
            buildMIFile(fp, content_output, msgno, msginfo.initoffset, msginfo.period, msginfo.offset, msginfo.jitter)
            res = runVeri(genVeriCmd(cmd, '', "query_msg.txt"))
            if res:
                print "\tProcess " + str(mpi.rank) + ": The initial message interface of MSG" +\
                      str(msgno + 1) + " is satisfied!"

                step = msginfo.jitter
                while step > 1 or res:
                    step = max(step / 2, 1)
                    print "\tProcess " + str(mpi.rank) + ": Binary search for best offset -- step = " + str(step)
                    buildMIFile(fp, content_output, msgno,
                                msginfo.initoffset, msginfo.period, msginfo.offset + step, msginfo.jitter - step)
                    res = runVeri(genVeriCmd(cmd, '', "query_msg.txt"))
                    if res:
                        msginfo.offset += step
                        msginfo.jitter -= step

                res = True
                step = msginfo.jitter
                while step > 1 or res:
                    step = max(step / 2, 1)
                    print "\tProcess " + str(mpi.rank) + ": Binary search for best jitter -- step = " + str(step)
                    buildMIFile(fp, content_output, msgno,
                                msginfo.initoffset, msginfo.period, msginfo.offset, msginfo.jitter - step)
                    res = runVeri(genVeriCmd(cmd, '', "query_msg.txt"))
                    if res:
                        msginfo.jitter -= step

                msglist.append(msginfo)
                res = True
                print "\tProcess " + str(mpi.rank) + ": The final message interface of MSG" + str(msgno + 1) + \
                      " is:\n" + "initialoffset: " + str(msginfo.initoffset) + " period: " + str(msginfo.period) + \
                      " offset: " + str(msginfo.offset) + " jitter: " + str(msginfo.jitter)
            else:
                print "\tProcess " + str(mpi.rank) + ": The initial message interface of MSG" + str(msgno) + \
                      " is NOT satisfied!"
                break

    # Compositional verification
    if res:
        outputMI(content_output, msglist)
        for i in range(partnum):
            buildCompVeriModelFile(fp, content_output, i)
            print "\tProcess " + str(mpi.rank) + ": Compositional MC verifies partition " + str(i + 1) + "..."
            res = runVeri(cmd)
            if res:
                schno += 1
                print "\tProcess " + str(mpi.rank) + ": Schedulability of partition " + str(i + 1) + " is satisfied!"
            else:
                print "\tProcess " + str(mpi.rank) + ": Schedulability of partition " + str(i + 1) + " is VIOLATED!"
                break

    return [res, schno]
