#!/usr/bin/env python
# -*- coding: utf-8 -*-

from math import log
#from itertools import accumulate
from random import uniform
from random import randrange
from random import sample
from random import triangular
from math import ceil

from GenSchedule import leastCommonMultiple
from GenSchedule import genSchedule


class GAIndividual(object):
    def __init__(self, ranges, encoding='binary', eps=0.001, strategy=None):
        '''
        Class for individual in population.

        :param ranges: value ranges for all entries in variants.
        :type ranges: list of range tuples. e.g. [(0, 1), (-1, 1)]

        :param encoding: gene encoding, 'decimal' or 'binary', default is binary.
        :param eps: decrete precision for binary encoding, default is 0.001.

        '''
        self._check_parameters()

        self.ranges = ranges
        self.eps = eps
        self.encoding = encoding
        self.strategy = strategy

        # Lengths for all binary sequence in chromsome.
        self.lengths = [int(ceil(log((b - a) / eps, 2))) for a, b in ranges]

        # Correct decrete precision according to binary sequence length.
        #self.precisions = [(b - a)/(2**l - 1) for l, (a, b) in zip(self.lengths, self.ranges)]
        self.precisions = [1 for i in range(len(ranges))]
        for i in range(0, len(ranges), 2): self.precisions[i] = self.eps

        # The start and end indices for each gene segment for entries in variants.
        self.gene_indices = self._get_gene_indices()

        # Generate randomly.
        self.variants = self._init_variants()

        # Gene encoding.
        self.chromsome = self.encode()

        # The number of generations that the individual lasts in elitist list
        self.lasting = 0

        # Is verified by symbolic model checking
        self.veri = False

    def init(self, chromsome=None, variants=None):
        '''
        Initialize the individual by providing chromsome or variants.
        If both chromsome and variants are provided, only the chromsome would
        be used.

        :param chromsome: chromesome sequence for the individual
        :type chromsome: list of float/int.

        :param variants: the variable vector of the target function.
        :type variants: list of float.
        '''
        if not any([chromsome, variants]):
            msg = 'Chromsome or variants must be supplied for individual initialization'
            raise ValueError(msg)

        if chromsome:
            self.chromsome = chromsome
            self.variants = self.decode()
        else:
            self.variants = variants
            self.chromsome = self.encode()

        self.strategy = None

        return self

    def clone(self):
        '''
        Clone a new individual from current one.
        '''
        indv = self.__class__(self.ranges, encoding=self.encoding, eps=self.eps)
        indv.init(variants=self.variants)

        return indv

    def _check_parameters(self):
        # Need implementation.
        pass

    def _init_variants(self):
        '''
        Initialize individual variants randomly.
        '''
        if self.eps >= 1:
            var = [0 for _ in self.ranges]
            num = len(self.ranges)

            scale = 100
            rate = [scale for _ in range(num / 2)]
            isolation = sample([i for i in range(1, scale)], num / 2 - 1)
            isolation.sort()
            isolation.insert(0, 0)
            isolation.append(scale)
            for j in range(len(isolation) - 1):
                rate[j] = isolation[j + 1] - isolation[j]

            startidx = randrange(0, num, 2)
            for i in range(0, num, 2):
                idx = (startidx + i) % num
                #prdi = randrange(self.ranges[idx][0], self.ranges[idx][1] + self.eps, int(self.eps))
                prdi = int(triangular(self.ranges[idx][0],
                                      self.ranges[idx][1],
                                      self.ranges[idx][1] / 2.0) / self.eps) * int(self.eps)
                budglb = self.ranges[idx + 1][0]
                budgub = int(float(prdi * rate[i / 2]) / scale)
                budgub = budgub if budgub > budglb else budglb + 1
                #budgi = randrange(budglb, budgub, int(self.precisions[idx + 1]))
                budgi = int(triangular(budglb, budgub, budgub) / self.precisions[idx + 1]) * int(self.precisions[idx + 1])
                #rate -= budgi * scale / prdi
                var[idx] = prdi
                var[idx + 1] = budgi
                pass
            #print str(startidx), var
            return var
        else:
            return [uniform(a, b) for a, b in self.ranges]

    def valid(self, MaxMajorFrame):
        '''
        Determine whether the current variant is valid.
        '''
        rate = 0.0
        partnum = len(self.variants) / 2
        valid = True

        # Check if the sum of processor utilization is less than 1
        for i in range(partnum):
            prdi = self.variants[2 * i]
            budgi = self.variants[2 * i + 1]
            prdbd = self.ranges[2 * i]
            budgbd = self.ranges[2 * i + 1]
            if budgi > prdi\
                or prdi > prdbd[1] or prdi < prdbd[0]\
                    or budgi > budgbd[1] or budgi < budgbd[0]:
                valid = False
                break
            rate += (float(budgi) / prdi)
            pass
        if rate > 1.0:
            valid = False

        # Check if a schedule can be built.
        prds = [self.variants[2 * i] for i in range(partnum)]
        budgs = [self.variants[2 * i + 1] for i in range(partnum)]

        # Solve major frame of ARINC 653 schedule
        majorframe = leastCommonMultiple(prds)
        if majorframe > MaxMajorFrame:
            valid = False
        else:
            # Index of the first partition window for each partition
            pindex = []
            # List of the partition windows of partition schedule
            plist = []
            # Generate ARINC 653 schedule from periodic partition parameter vector
            schpcnt = genSchedule(majorframe, prds, budgs, pindex, plist)
            if schpcnt < partnum:
                valid = False

        return valid

    def isEqual(self, indv):
        if not isinstance(indv, GAIndividual):
            return False
        else:
            return self.variants == indv.variants

    def setStrategy(self, strtgy):
        self.strategy = list(strtgy)

    def encode(self):
        '''
        Encode variant to gene sequence in individual using different encoding.
        '''
        if self.encoding == 'decimal':
            return self.variants

        chromsome = []
        for var, (a, _), length, eps in zip(self.variants, self.ranges,
                                            self.lengths, self.precisions):
            chromsome.extend(self.binarize(var - a, eps, length))

        return chromsome

    def decode(self):
        ''' 
        Decode gene sequence to variants of target function.
        '''
        if self.encoding == 'decimal':
            return self.variants

        variants = [self.decimalize(self.chromsome[start: end], eps, lower_bound)
                     for (start, end), (lower_bound, _), eps in
                     zip(self.gene_indices, self.ranges, self.precisions)]
        return variants

    def _get_gene_indices(self):
        '''
        Helper function to get gene slice indices.
        '''
        #end_indices = list(accumulate(self.lengths))
        end_indices = []
        accvl = 0
        for i in self.lengths:
            accvl += i
            end_indices.append(accvl)
            pass
        start_indices = [0] + end_indices[: -1]
        return list(zip(start_indices, end_indices))

    @staticmethod
    def binarize(decimal, eps, length):
        '''
        Helper function to convert a float to binary sequence.

        :param decimal: the decimal number to be converted.
        :param eps: the decrete precision of binary sequence.
        :param length: the length of binary sequence.
        '''
        n = int(decimal/eps)
        bin_str = '{:0>{}b}'.format(n, length)
        return [int(i) for i in bin_str]

    @staticmethod
    def decimalize(binary, eps, lower_bound):
        '''
        Helper function to convert a binary sequence back to decimal number.
        '''
        bin_str = ''.join([str(bit) for bit in binary])
        return lower_bound + int(bin_str, 2)*eps

