#!/usr/bin/env python
# -*- coding: utf-8 -*-

''' Genetic Algorithm engine definition '''

import logging
from sys import stdout
from random import shuffle

from .plugin_interfaces.analysis import OnTheFlyAnalysis
from .mpiutil import mpi


class GAEngine(object):
    '''
    Class for representing a Genetic Algorithm engine.
    '''
    def __init__(self, cfg, population, selection, crossover, mutation,
                 fitness=None, analysis=None):
        '''
        The Genetic Algorithm engine class is the central object in GAPY framework
        for running a genetic algorithm optimization. Once the population with
        individuals, a set of genetic operators and fitness function are setup,
        the engine object unites these informations and provide means for running
        a genetic algorithm optimization.

        :param population: The GAPopulation to be reproduced in evolution iteration.
        :param selection: The GASelection to be used for individual selection.
        :param crossover: The GACrossover to be used for individual crossover.
        :param mutation: The GAMutation to be used for individual mutation.
        :param fitness: The fitness calculation function for an individual in population.

        :param analysis: All analysis class for on-the-fly analysis.
        :type analysis: list of OnTheFlyAnalysis subclasses.
        '''
        # Check parameters validity.
        self._check_parameters()

        # Set logger.
        logger_name = 'ParamSweep.{}'.format(self.__class__.__name__)
        self.logger = logging.getLogger(logger_name)
        self.logger.setLevel(logging.DEBUG)
        self.logger.addHandler(logging.StreamHandler(stdout))

        # Attributes assignment.
        self.cfg = cfg
        self.population = population
        self.fitness = fitness
        self.selection = selection
        self.crossover = crossover
        self.mutation = mutation
        self.gi = 0

        self.analysis = [] if analysis is None else [a() for a in analysis]

        # Fitness dict
        self.FitDict = {}

        # Strategy selector -- select and refine the strategy parameters
        self.strategy_selection = None

    def run(self, ng=100):
        '''
        Run the Genetic Algorithm optimization iteration with specified parameters.

        :param control_parameters: An instance of GAControlParamters specifying
                                   number of evolution generations etc.
        '''
        if self.fitness is None:
            raise AttributeError('No fitness function in GA engine')

        # Setup analysis objects.
        for a in self.analysis:
            a.setup(ng=ng, engine=self)
        mpi.barrier()

        # Enter evolution iteration.
        for g in range(ng):
            self.gi = g
            # Scatter jobs to all processes.
            local_indvs = []
            local_size = mpi.split_size((self.population.size - self.population.elitist_size) // 2)
            msg = 'Processing Generation ' + str(g) + ' in Process ' + str(mpi.rank)
            self.logger.info(msg)

            # Calculate the fitness for the current generation distributedly
            self._distributed_fitness_calculation()
            mpi.barrier()

            # Run all analysis if needed.
            for a in self.analysis:
                if g % a.interval == 0:
                    a.register_step(g=g, population=self.population, engine=self)

            # Select the strategy parameters
            refinedstrtgy = self.strategy_selection.select(g, self.population, self.fitness) \
                if mpi.is_master else None
            refinedstrtgy = mpi.bcast_seq(refinedstrtgy)

            # Select the elitist individuals
            elitist = self.elitist_selection()
            # Set the strategy parameters to elitist individuals
            self.strategy_recombination(elitist, refinedstrtgy)

            # Fill the new population.
            for _ in range(local_size):
                # Select parents.
                parents = self.selection.select(self.population, fitness=self.fitness)

                # Give birth to children
                children_alive = []
                birth = 0
                max_birth = 50
                while len(children_alive) < 2:
                    # Crossover.
                    children = self.crossover.cross(*parents)
                    # Set the strategy parameters to children
                    self.strategy_recombination(children, refinedstrtgy)
                    # Mutation.
                    children = [self.mutation.mutate(child) for child in children]

                    birth += 1
                    if len(children_alive) < 2 and birth >= max_birth:
                        children_alive.extend(children)
                    else:
                        # Check if children are alive
                        for child in children:
                            if child.valid(self.cfg.MaxMajorFrame):
                                children_alive.append(child)

                # Collect children.
                local_indvs.extend(children_alive[:2])

            # Gather individuals from all processes.
            indvs = mpi.merge_seq(local_indvs)
            # The next generation.
            indvs.extend(elitist)
            shuffle(indvs)  # Keep load balancing among all processor cores
            self.population.individuals = indvs

        self.gi = ng

        # Calculate the fitness for the last generation distributedly
        self._distributed_fitness_calculation()

        # Perform the analysis post processing.
        for a in self.analysis:
            a.finalize(population=self.population, engine=self)

    def _distributed_fitness_calculation(self):
        '''
        Split the fitness calculation over all the processes and merge results to the master process.
        '''
        popldata = [self.population[i].variants for i in range(self.population.size)] if mpi.is_master else None
        strtgydata = [self.population[i].strategy for i in range(self.population.size)] if mpi.is_master else None
        lastingdata = [self.population[i].lasting for i in range(self.population.size)] if mpi.is_master else None

        # Broadcast the population in master process to all processes.
        popldata = mpi.bcast_seq(popldata)
        strtgydata = mpi.bcast_seq(strtgydata)
        lastingdata = mpi.bcast_seq(lastingdata)
        if not mpi.is_master:
            for i in range(self.population.size):
                self.population[i].init(variants=popldata[i])
                self.population[i].setStrategy(strtgydata[i])
                self.population[i].lasting = lastingdata[i]

        # Split fitness calculation to all processes.
        splited_indvs = mpi.split_seq(self.population.individuals)
        local_dict = {}
        for indv in splited_indvs:
            local_dict[tuple(indv.variants)] = [self.fitness(indv), indv.veri]

        # Gather the fitness dictionaries from all processes.
        newdicts = dict(mpi.merge_seq([[k, v] for k, v in local_dict.items()]))
        self.FitDict.update(newdicts)

        msg = 'Fitness calculation completed in Process ' + str(mpi.rank)
        self.logger.info(msg)

    def _check_parameters(self):
        '''
        Helper function to check parameters of engine.
        '''
        # Need implementation.
        pass

    # Decorators.

    def fitness_register(self, fn):
        '''
        A decorator for fitness function register.
        '''
        self.fitness = fn

    def analysis_register(self, analysis_cls):
        '''
        A decorator for analysis regsiter.
        '''
        if not issubclass(analysis_cls, OnTheFlyAnalysis):
            raise TypeError('analysis class must be subclass of OnTheFlyAnalysis')

        # Add analysis instance to engine.
        analysis = analysis_cls()
        self.analysis.append(analysis)

    def strategy_selection_register(self, obj):
        '''
        A decorator for strategy refiner function register.
        '''
        self.strategy_selection = obj

    def strategy_recombination(self, indvs, strtgy):
        '''
        Set refined strategy parameters into the selected parents.
        '''
        for indv in indvs:
            indv.setStrategy(strtgy)

    def elitist_selection(self):
        '''
        Select elitist individuals from the current population.
        '''
        sortedindvs = sorted(self.population.individuals, key=self.fitness, reverse=True)
        sortedindvs = sortedindvs[:self.population.elitist_size]

        for i in range(len(sortedindvs)):
            sortedindvs[i].lasting += 1

        return sortedindvs
