#!/usr/bin/env python
# -*- coding: utf-8 -*-

from random import uniform

from ...plugin_interfaces.operators.crossover import GACrossover


class IntermediateRecombination(GACrossover):
    def __init__(self, d=0.25):
        '''
        Crossover operator with intermediate recombination algorithm,
        see https://en.wikipedia.org/wiki/Crossover_(genetic_algorithm)

        :param d: a scaling factor will be chosen uniformly at random over an interval [-d, 1+d]
        :type d: float in (0.0, 1.0]
        '''

        if d <= 0.0 or d > 1.0:
            raise ValueError('Invalid d value in intermediate recombination!')

        self.d = d

    def cross(self, father, mother):
        '''
        Cross chromsomes of parent using intermediate recombination method.
        '''
        # Copy the variants of parents
        vars1 = list(father.variants)
        vars2 = list(mother.variants)

        for i, (g1, g2) in enumerate(zip(vars1, vars2)):
            ai = uniform(-self.d, 1 + self.d)
            vars1[i] = g1 + int(round(ai * (g2 - g1) / father.precisions[i])) * father.precisions[i]
            if vars1[i] < father.ranges[i][0]:
                vars1[i] = father.ranges[i][0]
            if vars1[i] > father.ranges[i][1]:
                vars1[i] = father.ranges[i][1]

        child = father.clone()
        child.init(variants=vars1)

        return [child]
