#!/usr/bin/env python
# -*- coding: utf-8 -*-

from random import random
from random import randrange

from ...plugin_interfaces.operators.crossover import GACrossover


class RadianLineRecombination(GACrossover):
    def __init__(self, d=0.25, step=10):
        '''
        Crossover operator with randian line recombination algorithm,
        see https://en.wikipedia.org/wiki/Crossover_(genetic_algorithm)

        :param d: a scaling factor will be chosen uniformly at random over an interval [-d, 1+d]
        :type d: float in (0.0, 1.0]

        :param step: the basic unit of variables in integer
        '''

        if d <= 0.0 or d > 1.0:
            raise ValueError('Invalid d value in radian line recombination!')
        if not isinstance(step, int):
            raise ValueError('Invalid mutation step')

        self.d = d
        self.step = step

    def cross(self, father, mother):
        '''
        Cross chromsomes of parent using radian line recombination method.
        '''
        # Copy the variants of parents
        vars1 = list(father.variants)
        vars2 = list(mother.variants)

        ai = 0
        crossvar = [0.0, 0.0]
        for i, (g1, g2) in enumerate(zip(vars1, vars2)):
            # Update factor for each partition
            if i % 2 == 0:
                ai = random()
                crossvar[0] = g1 + ai * (g2 - g1)
                vars1[i] = randrange(father.ranges[i][0], father.ranges[i][1], self.step)
            else:
                crossvar[1] = g1 + ai * (g2 - g1)
                vars1[i] = int(round(vars1[i - 1] * crossvar[1] / crossvar[0] / self.step)) * self.step

            if vars1[i] < father.ranges[i][0]:
                vars1[i] = father.ranges[i][0]
            if vars1[i] > father.ranges[i][1]:
                vars1[i] = father.ranges[i][1]

        child = father.clone()
        child.init(variants=vars1)

        return [child]
