#!/usr/bin/env python
# -*- coding: utf-8 -*-

from random import random
from random import gauss
from math import e
import numpy as np


from ...plugin_interfaces.operators.mutation import GAMutation


class RotatingGaussMutation(GAMutation):
    def __init__(self, pm, lrate, sranges=()):
        '''
        Mutation operator with rotating gauss mutation implementation.

        :param pm: The probability of mutation (usually between 0.001 ~ 0.1)
        :type pm: float in (0.0, 1.0]

        :param lrate: the learning rate of covariance matrix of the mutating normal distribution
        '''
        if pm < 0.0 or pm > 1.0:
            raise ValueError('Invalid mutation probability in rotating gauss mutation!')
        if lrate <= 0.0 or lrate > 1.0:
            raise ValueError('Invalid learning rate in rotating gauss mutation!')

        self.pm = pm
        self.lrate = lrate
        # Ranges of strategy parameters[(min1, max1),...]
        self.sranges = sranges

    def mutate(self, individual):
        '''
        Mutate the individual.
        '''
        for i, var in enumerate(individual.variants):
            if i % 2 == 0:
                x = float(var)
            else:
                y = float(var)

                # Mutate the strategy parameters
                sigma_1 = individual.strategy[i - 1] * e**(self.lrate * gauss(0, 1))
                sigma_2 = individual.strategy[i] * e**(self.lrate * gauss(0, 1))
                individual.strategy[i - 1] = sigma_1
                individual.strategy[i] = sigma_2

                # Build the covariance matrix of the mutating normal distribution
                sintheta = y / (x**2 + y**2)**0.5
                costheta = x / (x**2 + y**2)**0.5
                cov = [[(sigma_1 * costheta)**2 + (sigma_2 * sintheta)**2,
                        (sigma_1**2 - sigma_2**2) * sintheta * costheta],
                       [(sigma_1**2 - sigma_2**2) * sintheta * costheta,
                        (sigma_1 * sintheta)**2 + (sigma_2 * costheta)**2]
                       ]

                # Mutate the object parameters on the slope line
                z = (x**2 + y**2)**0.5
                zmim = individual.ranges[i - 1][0] / costheta
                if z < 2 * sigma_1 + zmim:
                    xp = (2 * sigma_1 + zmim) * costheta
                    yp = (2 * sigma_1 + zmim) * sintheta
                else:
                    xp = x
                    yp = y

                mv = False
                maxgen = 50
                while maxgen > 0 and not mv:
                    # Mutate the object parameters on the rotated normal distribution
                    [deltax, deltay] = np.random.multivariate_normal([0, 0], cov)

                    x = xp + deltax
                    y = yp + deltay

                    # Align with the grid
                    x = int(round((x - individual.ranges[i - 1][0]) / individual.precisions[i - 1])) * \
                        individual.precisions[i - 1] + individual.ranges[i - 1][0]
                    y = int(round((y - individual.ranges[i][0]) / individual.precisions[i])) * \
                        individual.precisions[i] + individual.ranges[i][0]

                    # Check the range of parameters
                    if individual.ranges[i - 1][0] <= x <= individual.ranges[i - 1][1] and\
                        individual.ranges[i][0] <= y <= individual.ranges[i][1]:
                        mv = True

                    maxgen -= 1

                # Ensure the range of parameters
                if x < individual.ranges[i - 1][0]:
                    x = individual.ranges[i - 1][0]
                if x > individual.ranges[i - 1][1]:
                    x = individual.ranges[i - 1][1]
                individual.variants[i - 1] = x

                if y < individual.ranges[i][0]:
                    y = individual.ranges[i][0]
                if y > individual.ranges[i][1]:
                    y = individual.ranges[i][1]
                individual.variants[i] = y

        # Update variants.
        individual.chromsome = individual.encode()

        return individual
